# -*- coding: utf-8 -*-
"""
Created on Wed Aug 13:34:59 2015

@author: nikol_000
"""

'''
Pseudocode for the MPC-algorithm

position = init_pos
waypoint = init_waypoint

plant_input, opt_time = optimal_controller(position, waypoint)

while icebergs
    waypoint = find_current_waypoint(position)
    
    next_plant_input, opt_time = optimal_controller(position, waypoint)
    
    new_position = plant_model_simulation(position, plant_input, opt_time)
        
        
    save_data_for_graphs()
    
    position = new_position
    plant_input = next_plant_input
    total_time += opt_time
        
    

'''

from casadi import *
from plot import *
import matplotlib.pyplot as plt
import time
import random
import pylab
import numpy as NP
import cmath as cmath
import matplotlib.pyplot as plt

################################################
########Global constants and functions##########
################################################
random.seed(65)

###Modes###
#WIND
WIND = True     
WIND_CHANGE = False
WIND_COMPENSATING = True
WIND_ESTIMATION = False
#Simluation
FIXED_TIME = True


####################

icebergs_init = [ [0,750], [1000, 600], [1300, 1500], [350, 1600], [500, 2500], [-1000, 2500] ]

x_init = [0,0,pi/4]

if WIND:
#    true_wind = [10, pi/4]
    true_wind = [5, pi/4]
else:
    true_wind = [0,0]

if WIND_COMPENSATING:
#    est_wind = [10, pi/4]
    est_wind = [5, pi/4]
else:
    est_wind = [0,0]


################################################
####      Optimal UAV controller        ########
################################################

UAV_speed = 22
g0 = 9.81
u_min = -g0/UAV_speed*tan(5*pi/36)
u_max =  g0/UAV_speed*tan(5*pi/36)


turn_radius = ( (2*pi/u_max)*UAV_speed ) / (2*pi)

#Optimizing model
x = SX.sym('x', 3) #states x, y, psi
u = SX.sym('u') #control u

#ODE right hand side

x_dot = vertcat([   UAV_speed*NP.cos(x[2]) + est_wind[0]*NP.cos(est_wind[1]) , \
                    UAV_speed*NP.sin(x[2]) + est_wind[0]*NP.sin(est_wind[1]), \
                    u ] )
f_opt = SXFunction([ x, u], [x_dot] )


#Objective function

#Control bounds
u_lb = NP.array([u_min])
u_ub = NP.array([u_max])
u_init = NP.array([0])

#Initialize functions
f_opt.init()

# Dimensions
nx = 3
nu = 1

#Collocation Coefficients
d = 3
tau_root = collocationPoints(d,"legendre")

# Dimensionless time inside one control interval
tau = SX.sym("tau")

# Calculate coefficients
C = DMatrix.nan(d+1,d+1)  # For the collocation equation
D = DMatrix.nan(d+1)      # For the continuity equation
for j in range(d+1):
  # Lagrange polynomial
  L = 1
  for r in range(d+1):
    if r != j:
      L *= (tau-tau_root[r])/(tau_root[j]-tau_root[r])

  # Evaluate the polynomial at the final time to get the coefficients of the 
  # continuity equation
  lfcn = SXFunction([tau],[L])
  lfcn.init()
  [D[j]] = lfcn([1.0])

  # Evaluate the time derivative of the polynomial at all collocation points to
  # get the coefficients of the collocation equation
  dL = gradient(L,tau)
  dfcn = SXFunction([tau],[dL])
  dfcn.init()
  for r in range(d+1):
    [C[j,r]] = dfcn([tau_root[r]])
    
################################################

#Optimization algorithm with time bisection

def wind_componenst(w, position, target ):
    x_wind = w[0]*NP.cos(w[1])
    y_wind = w[0]*NP.sin(w[1])
    wind_vec = NP.array([x_wind, y_wind])
    
    distance_vec = NP.array([target[0] - position[0], target[1] - position[1]])
    unit_dist_vec = distance_vec / sum(distance_vec*distance_vec)**0.5

    alongWind_vec = NP.dot(wind_vec, unit_dist_vec)*unit_dist_vec
    crossWind_vec = wind_vec - alongWind_vec


    alongWind = NP.dot(alongWind_vec, unit_dist_vec)
    crossWind = NP.dot(crossWind_vec, wind_vec/sum(wind_vec*wind_vec)**0.5 )
    
    return alongWind, crossWind_vec
    

def optimal_controller( position, targets):
    
    icebergs = targets
    x_target = targets
    opt_time_start = time.time()
    
    simdata = [] #saving simulation data for endprint
    simtime = []
    x_out = []
    
    n_wp = len(targets)
    x_accuracy = NP.array( [10, 10, pi/72] )
    u_init = 0


    
#    finish_angle = pi/2
#    print "icebergs", icebergs
#    
    #Generating targets
#    target_angles = []
#    for i in range(n_wp-1):
#        if i == 0:
#            current_straight_heading = x_init[2]
#        else:
#            current_straight_heading = NP.arctan2( (icebergs[i][1] - icebergs[i-1][1]), (icebergs[i][0] - icebergs[i-1][0]) )            
#        next_straight_heading = NP.arctan2( (icebergs[i+1][1] - icebergs[i][1]), (icebergs[i+1][0] - icebergs[i][0]) )     
#        target_angles.append( ( current_straight_heading + next_straight_heading ) /2 )
#        
#    target_angles.append(finish_angle)    
#    for i in range(n_wp):
#        x_target.append( icebergs[i] )
#        x_target[i].append(target_angles[i])         
#      
#    print "waypoints before loop", waypoints
#    print "icebergs before loop", icebergs       
#    print "x_target before loop", x_target
    #Optimization loop for the icebergs

    for i, iceberg in enumerate(icebergs):

        subsim = [] # saving simulation data per iceberg
        
        #Calculating an approximating for the time needed to get to the point based on distance and heading difference
        distance = sqrt( (x_target[i][0] - position[0])**2 + (x_target[i][1] - position[1])**2 )
               
        along_wind, cross_wind = wind_componenst(est_wind, position, x_target[i])
                    
        dist_time = distance.item()/(UAV_speed+along_wind)
        
        headingDifference =  abs(x_target[i][2] - position[2] )
        
    #    if i < len(icebergs) - 2:
    #        nextHeadingDiff = x_target[i+1][2] - x_target[i][2]
    #        newheading = x_target[i][2] - ( nextHeadingDiff )*turn_radius/distance    
    
            
        #Control intervals 
        n = int(dist_time)  #control intervals dependent on distance)
        
        # State bounds and initial guess
        x_min =  [-inf, -inf, -inf]
        x_max =  [ inf,  inf,  inf]
        xi_min = position 
        xi_max = position
        #xf_min = [ -inf,  -inf, -inf]
        #xf_max = [ inf,  inf,  inf]

        xf_min = x_target[i] - x_accuracy
        xf_max = x_target[i] + x_accuracy
        if i == len(icebergs)-1:
            xf_min[2] = x_target[i][2] - pi/2
            xf_max[2] = x_target[i][2] + pi/2
        
        #Total number of variables
        NX = n*(d+1)*nx + nx    #collocation states
        NU = n * nu             #controls
        NW = NX+ NU             #states + contorl + end time
        #NLP vector
        w = MX.sym('w', NW)
        
        
        vars_lb = NP.zeros(NW)
        vars_ub = NP.zeros(NW)
        vars_init = NP.zeros(NW)
        offset = 0
        
        X = NP.resize(NP.array([], dtype=MX), (n+1, d+1))
        U = NP.resize(NP.array([], dtype=MX), n )
        
        #Collocated states and parametrized controls
        for k in range(n):
            #collocated states
            for j in range(d+1):
                # Get the expression for the state vector
                X[k, j] = w[offset:offset+nx]
                
                vars_init[offset:offset+nx] = position
                
                if k ==0 and j ==0:
                    vars_lb[offset:offset+nx] = xi_min
                    vars_ub[offset:offset+nx] = xi_max
                else:
                    vars_lb[offset:offset+nx] = x_min
                    vars_ub[offset:offset+nx] = x_max
                offset += nx
            #Parametrized controls
            U[k] = w[offset:offset+nu]
            vars_lb[offset:offset+nu] = u_min
            vars_ub[offset:offset+nu] = u_max
            vars_init[offset:offset+nu] = u_init
            offset += nu
        #state at end time
        X[n, 0] = w[offset:offset+nx]
        vars_lb[offset:offset+nx] = xf_min
        vars_ub[offset:offset+nx] = xf_max
        vars_init[offset:offset+nx] = position
        offset += nx
        
        
        #Objective functions  
        obj_term1 = [ (x[0]-x_target[i][0])**2 + (x[1]-x_target[i][1])**2 + (x[2]-x_target[i][2])**2   ]
        obj_func1 = SXFunction( [ x ], obj_term1 )
        obj_func1.init()
        
        obj_term2 = [ u*u ]
        obj_func2 = SXFunction( [x,u], obj_term2)
        obj_func2.init()
        
        #Objective function to minimize positin error at end time
        Obj1 = 0
        for j in range(1,d+1):
            obj = obj_func1.call( [ X[n-1,j] ] )[0]
            Obj1 += obj
        #Objective function to minimize turning over the interval
        Obj2 = 0
        for k in range(0, n):
            for j in range(1,d+1):
                obj = obj_func2.call( [X[k,j], U[k] ] )[0]
                Obj2 += obj
        Obj = Obj1 + Obj2
    
        #exit criteria bools
        iteration = 0
        finished= False
        force_finish = False
        last_fail = -inf
        last_succ = inf
        last_msg = None
        t_curve = 3
        t_const = 0.0
        
        
        while not finished: #Bisection loop to optimize end time
                    
            tf = dist_time + t_const + t_curve*abs(headingDifference) 
            print tf
             
            DT = tf/n        
            
            #Constraint function for the NLP
            g = []
            lbg = []
            ubg = []
               
            # For all finite elements
            for k in range(n):
              
              # For all collocation points
              for j in range(1,d+1):
                    
                # Get an expression for the state derivative at the collocation point
                xp_jk = 0
                for r in range (d+1):
                  xp_jk += C[r,j]*X[k,r]
                  
                # Add collocation equations to the NLP
                [fk] = f_opt.call( [ X[k,j], U[k] ] )
                g.append(DT*fk - xp_jk )
                lbg.append(NP.zeros(nx)) # equality constraints
                ubg.append(NP.zeros(nx)) # equality constraints
            
              # Get an expression for the state at the end of the finite element
              xf_k = 0
              for r in range(d+1):
                xf_k += D[r]*X[k,r]
            
              # Add continuity equation to NLP
              g.append(X[k+1,0] - xf_k)
              lbg.append(NP.zeros(nx))
              ubg.append(NP.zeros(nx))
            
            #consentrate constraints
            g = vertcat(g)
        
        
            #NLP
            nlp = MXFunction(nlpIn(x=w), nlpOut(f=Obj, g=g) )
            
            solver = NlpSolver("ipopt", nlp)
            solver.setOption("expand", True)
            solver.setOption("linear_solver", "ma27")
            solver.setOption('verbose', False)
            solver.setOption("max_iter",1000)
            solver.init()
            
            solver.setInput(vars_init, 'x0')
            solver.setInput(vars_lb, 'lbx')
            solver.setInput(vars_ub, 'ubx')
            solver.setInput(NP.concatenate(lbg), 'lbg')
            solver.setInput(NP.concatenate(ubg), 'ubg')
            
            solver.evaluate()
            
            #Checking if solution is satisfactory and updating exit criteria
            exit_msg = solver.getStat('return_status')
            solver.getStat('t_mainloop')
            
            subsim.append( [iteration, t_const, exit_msg ] )  #saving data for end print
            
            if force_finish:
                finished = True
            elif exit_msg == "Solve_Succeeded":
                last_succ = t_const
                if last_fail != -inf:
                    t_const = (last_fail + t_const) / 2 
                else:
                    t_const -= 3      
            elif exit_msg == "Infeasible_Problem_Detected":
                last_fail = t_const
                if last_succ != inf:
                    t_const = (last_succ + t_const) / 2
                else:
                    t_const += 3
            elif exit_msg == "Invalid_Number_Detected":
                last_fail = t_const
                if last_succ != inf:
                    t_const = (last_succ + t_const) / 2
                else:
                    t_const += 3
            elif exit_msg == "Maximum_Iterations_Exceeded":
                last_fail = t_const
                if last_succ != inf:
                    t_const = (last_succ + t_const) / 2
                elif last_succ == t_const:
                    t_const += 2
                else:
                    t_const += 2
            elif exit_msg == "Restoration_Failed":
                t_const -= 1             
                if last_msg == "Restoration Failed":
                    t_const += 2            
                    
            print "t_const: " + str( t_const )
            print "last_fail: " +  str(last_fail)
            print "last_succ: "+ str( last_succ)
           
            if exit_msg == "Solve_Succeeded":
                simOut = solver.getOutput('x')
                #bisection_graphs.append( simOut )
                    
            last_msg = exit_msg
            iteration += 1
            if ( iteration >= 5 and exit_msg == "Solve_Succeeded") or iteration > 5:
                if iteration > 5:
                    t_const = last_succ
                    force_finish = True
                else:
                    finished = True
                
        
        simtime.append( tf )
        
        simdata.append( ["Iceberg {0}".format(i+1) , subsim ])
        
        simOut = solver.getOutput('x')
        x_out.append( simOut )
        
        position = NP.concatenate( NP.array( simOut[-3:] ) ) #Setting the endpoint as the next starting point
        u_init = NP.concatenate( NP.array( simOut[-4:-3] ) )    

    opt_time_end = time.time()
    opt_time_used = opt_time_end - opt_time_start
    
    trajectory_time = sum(simtime)
    
    optimal_positions = NP.array([ NP.concatenate( [NP.array( x_out[i][0::(d+1)*nx+nu] ) for i in range(n_wp)]  ),
                        NP.concatenate( [NP.array( x_out[i][1::(d+1)*nx+nu] ) for i in range(n_wp)]  ),
                        NP.concatenate( [NP.array( x_out[i][2::(d+1)*nx+nu] ) for i in range(n_wp)]  ) ] )
    optimal_inputs = NP.concatenate( [ x_out[i][(d+1)*nx::(d+1)*nx+nu] for i in range(n_wp)] )        
    
    #removing target angle from icebergs (dunno why they are there in the first place...)
    for i in range(n_wp):
        waypoints[i] = waypoints[i][0:2]
        icebergs[i] = icebergs[i][0:2]
    for data in simdata:
        print data
        
    return  optimal_inputs, optimal_positions, opt_time_used, trajectory_time


################################################
#####         Plant UAV model            #######
################################################
T = 1
N = 1
turn_delay = u_max*2

def rk4(x0, u, w, T, N, f):
    x = x0
    DT = T/N
    for k in range(N):
        k1 = DT*f(x, u, w)
        k2 = DT*f(x + 0.5*k1, u, w )
        k3 = DT*f(x + 0.5*k2, u, w )
        k4 = DT*f(x + k3,u, w)
        x = x + (k1 + 2*k2 + 2*k3 +k4)/6
            
    return [round(x[0], 3), round(x[1], 3), round(x[2], 5) ]

def plant_UAV(x, u, w):
    X = NP.array([      UAV_speed*NP.cos(x[2]) + w[0]*cos(w[1]) ,  \
                        UAV_speed*NP.sin(x[2]) + w[0]*sin(w[1]) ,  \
                        u ])
    return X                        

def plant_model_sim(position, plant_input, true_wind, opt_time, trajec_time, last_opt_time=0):
    x = []
    x_trajec = []
    x_trajec.append(position) 
    u_step_length = trajec_time/len(plant_input)
    
    if last_opt_time != 0:
        starting_input_step = int(round(last_opt_time/u_step_length, 0) )
    else:
        starting_input_step = 0
      
    #Using simulation step lenght of roughly 0.5s
    if FIXED_TIME:
        simulation_steps = int(ceil(5/u_step_length))
    else:
        simulation_steps = int( ceil(opt_time/u_step_length) )
    t_simul = simulation_steps*u_step_length
    
    T = u_step_length
    N = 1
    x0 = position
    w = true_wind
       
    for i in range(simulation_steps):        
        
        input_step = int(opt_time/simulation_steps*(i+1) )   
        
        u = plant_input[i+starting_input_step]
        x = rk4(x0 ,u ,w ,T ,N , plant_UAV )
        x_trajec.append(x)
        x0 = x
     
    return x, x_trajec, t_simul
    

##### Wind estimator #####
def wind_estimator(position, optimal_trajec, t_simul, t_trajec, est_wind ):
    
    current_step = t_simul*len(optimal_trajec)/t_trajec
    optimal_position = [optimal_trajec[0][int(ceil(current_step))]+optimal_trajec[0][int(current_step)] / 2, \
                        optimal_trajec[1][int(ceil(current_step))]+optimal_trajec[1][int(current_step)] / 2 ]
    pos_diff = [optimal_position[0] - position[0], optimal_position[1] -  position[1]]

    
    est_wind_drift = [ est_wind[0]*t_simul*NP.cos(est_wind[1]), est_wind[0]*t_simul*NP.sin(est_wind[1]) ]
    total_wind_drift = [ est_wind_drift[0] + pos_diff[0], est_wind_drift[1] + pos_diff[1] ]
     
    magnitude = sqrt( total_wind_drift[0]*total_wind_drift[0] + total_wind_drift[1]*total_wind_drift[1] )/t_simul
    direction = NP.arctan2(total_wind_drift[1], total_wind_drift[0])
        
    return [magnitude, direction] 
    
#def kalman_filter():
    
                  
##### Real Wind Update #####
def update_real_wind(wind, t_simul):
    min_magnitude = 4    
    max_magnitude = 12
       
    max_angle_change = pi/180
    max_magnutude_change = 0.05
    
    magnitude = wind[0] + random.uniform(-1,1)*max_magnutude_change*t_simul
    if magnitude > max_magnitude:
        magnitude = max_magnitude
    elif magnitude < min_magnitude:
        magnitude = min_magnitude
        
    direction = wind[1] + random.uniform(-1,1)*max_angle_change*t_simul
    
    return [magnitude, direction]
    
##### saving data for graphs #####

opt_trajec_data = []
plant_trajec_data = []

def save_graph_data(opt_trajec, plant_trajec):
    opt_trajec_data.append(opt_trajec)
    plant_trajec_data.append(plant_trajec)

def ploteverything():
    plotall_opt_trajec(opt_trajec_data)
    plotall_plant_trajec(plant_trajec_data)
    plot_icebergs(icebergs_init)
    grid(b=True)


################################################
#####            Main MPC code           #######
################################################


waypoints = icebergs_init[::]
position = x_init
visited_waypoints = []





last_opt_time = 0


finished = False
total_time = 0

innom = 0
kommet_hit = False

#Generating waypoints target angles
x_target = []
finish_angle = NP.arctan2( (waypoints[-1][1] - waypoints[-2][1]), (waypoints[-1][0] - waypoints[-2][0]) )
#Generating targets
target_angles = []
for i in range(len(waypoints)-1):
    if i == 0:
        current_straight_heading = x_init[2]
    else:
        current_straight_heading = NP.arctan2( (waypoints[i][1] - waypoints[i-1][1]), (waypoints[i][0] - waypoints[i-1][0]) )            
    next_straight_heading = NP.arctan2( (waypoints[i+1][1] - waypoints[i][1]), (waypoints[i+1][0] - waypoints[i][0]) )     
    target_angles.append( ( current_straight_heading + next_straight_heading ) /2 )
    
target_angles.append(finish_angle)    
for i in range(len(waypoints)):
    x_target.append( waypoints[i] )
    x_target[i].append(target_angles[i]) 

#initial Optimization
opt_targets = x_target[0:3]
#opt_targets = x_target
initial_optimization = optimal_controller(position, opt_targets)
plant_input = initial_optimization[0]
last_opt_result = initial_optimization

#
est_wind_test = [ est_wind ]
while not finished:
        
    opt_targets = x_target[0:3]
    
    #Running optimization to find the next plant inputs
    next_plant_input, optimal_trajectory,  opt_time, trajec_time = optimal_controller(position, opt_targets)

    #Running simulation with the previous plant inputs for as long time as the optimization took
    
    #if one of the optimization fails
    if trajec_time == inf:
        
        new_position, plant_trajectory, t_simul = plant_model_sim(position, plant_input, true_wind, opt_time, last_succ_opt[3], last_opt_time=last_succ_opt[4])
        
        last_succ_opt[4] += t_simul
            
        if WIND_ESTIMATION:
            est_wind_test.append( wind_estimator(new_position, optimal_trajectory, t_simul, trajec_time, est_wind_test[-1]) )
    #Regular case    
    else:
        new_position, plant_trajectory, t_simul = plant_model_sim(position, next_plant_input, true_wind, opt_time, trajec_time)
        
        if WIND_ESTIMATION:
            est_wind_test.append( wind_estimator(new_position, optimal_trajectory, t_simul, trajec_time, est_wind_test[-1]) )
        
        #Save backup result in case the optmization doesn't find a solution
        last_succ_opt = [next_plant_input, optimal_trajectory, opt_time, trajec_time, t_simul]


    #Updating wind
    if WIND_CHANGE:
        true_wind = update_real_wind(true_wind, t_simul)
    
    #saving data for graphs
            
    save_graph_data(optimal_trajectory, plant_trajectory)
    
    kommet_hit = True
    
    #Checking if one waypoint is reached.
    waypoint_reached = False
    for i in range(len(plant_trajectory)):    
        waypoint_distance = sqrt( (waypoints[0][0] - plant_trajectory[i][0])**2 + (waypoints[0][1] - plant_trajectory[i][1])**2 )
        print waypoint_distance
        if waypoint_distance < 50:
            waypoint_reached = True
            print "Waypoint reached"
    if waypoint_reached:
        waypoints = waypoints[1::]
        x_target = x_target[1::]
    if not waypoints:
        finished = True        
    innom += 1     
    if innom > 100:
        finished = True
    #updating variables for next loop
    
    
    last_opt_time = t_simul
    position = new_position
    plant_input = next_plant_input
    total_time += opt_time
    
    
    











